From 0101278b67b64288c8d4afb282e0d4deb632980e Mon Sep 17 00:00:00 2001
Message-ID: <0101278b67b64288c8d4afb282e0d4deb632980e.1749546774.git.edwin.torok@cloud.com>
In-Reply-To: <cover.1749546774.git.edwin.torok@cloud.com>
References: <cover.1749546774.git.edwin.torok@cloud.com>
From: =?UTF-8?q?Edwin=20T=C3=B6r=C3=B6k?= <edwin.torok@cloud.com>
Date: Tue, 10 Jun 2025 09:08:46 +0100
Subject: [PATCH 1/3] openssl: enable readahead
MIME-Version: 1.0
Content-Type: text/plain; charset=UTF-8
Content-Transfer-Encoding: 8bit

`strace` on `stunnel` was showing repeated syscalls for 5 bytes, followed by
another for the actual data:
```
[pid 43497] read(10, "\27\3\3@\30", 5)  = 5
[pid 43497] read(10, "\34uz\323?\223SL\316K)\34[\260\263\333l\377\23\323\267i\237\263\365\16\3\241\376\27\32M"..., 16408) = 16408
```

(e.g. OpenSSL 3.2.4 on Fedora42)

This is because readahead is not enabled on the OpenSSL context (I'm assuming for
backwards compatibility).

According to the manpage at https://docs.openssl.org/3.3/man3/SSL_CTX_set_read_ahead/#notes:
> If the application wants to continue to use the underlying transport (e.g. TCP connection)
> after the SSL connection is finished using SSL_shutdown() reading ahead should be turned off.
> Otherwise the SSL structure might read data that it shouldn't

However I'm not aware of any protocols (that `stunnel` would support) which
needs this. STARTTLS is the only one that would combine unencrypted and
encrypted traffic on the same socket, but the switch is one way: once it
entered TLS mode, it doesn't go back.

Enable readahead by default on the OpenSSL context, which cuts the amount of read syscalls into half:
```
[pid 43828] read(10, "\322\0354.c}\373\307'\"\30,\253=\257\0\220\376\236\240Z\216\322$\300\276\306\333\254TE{"..., 16113) = 16113
[pid 43828] poll([{fd=6, events=POLLIN|POLLRDHUP}, {fd=3, events=POLLIN|POLLOUT|POLLRDHUP}, {fd=10, events=POLLIN|POLLRDHUP}], 3, 43200000) = 2 ([{fd=3, revent
s=POLLOUT}, {fd=10, revents=POLLIN}])
[pid 43828] write(3, "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"..., 16384) = 16384
[pid 43828] read(10, "\27\3\3@\30\236\6\366\v\365\243`*\305\26\307\266\3315#\250k\357\30\177\220\177\32\311\347\312\204"..., 16713) = 16713
```

This also reduces HTTPS download times by 17%, as confirmed by `ministat`:
```
x tmp-baseline/stunnel_client_nginx.dat
+ tmp-readahead/stunnel_client_nginx.dat
    N           Min           Max        Median           Avg        Stddev
x  30      9.886741     10.776708     10.581105     10.528033     0.1868863
+  30      8.472009      9.317522      8.623337     8.6852521    0.21674743
Difference at 95.0% confidence
        -1.84278 +/- 0.104607
        -17.5036% +/- 0.923458%
        (Student's t, pooled s = 0.202368)
```

This function got introduced in the following OpenSSL commit:
OpenSSL 0.9.2b: 413c4f45ed ("Updates to the new SSL compression code")

stunnel versions 5.71 and 5.72 actually had this enabled, but 5.73
removed it again.

It may expose latent bugs in various protocol implementations (e.g. it
exposed latent bugs in curl's FTPS and IMAPS code).

Perhaps this should be configurable by the application, at least for
HTTPS.

Signed-off-by: Edwin Török <edwin.torok@cloud.com>
---
 src/ctx.c | 2 ++
 1 file changed, 2 insertions(+)

diff --git a/src/ctx.c b/src/ctx.c
index a913905..65f7f19 100644
--- a/src/ctx.c
+++ b/src/ctx.c
@@ -183,6 +183,8 @@ int context_init(SERVICE_OPTIONS *section) { /* init TLS context */
         return 1; /* FAILED */
     }
 
+    SSL_CTX_set_read_ahead(section->ctx, 1);
+
     /* set supported protocol versions */
 #if OPENSSL_VERSION_NUMBER>=0x10100000L
     if(section->min_proto_version &&
-- 
2.43.5

